			include "propeller.i"
			include "demo.i"

			org	0

			mov	t1, #DITHER_MANDEL
			wrbyte	t1, =@ditherstyle

loop
			' Fetch a new job (line to render).

:lock			lockset	zero			wc
		if_c	jmp	#:lock
			rdword	ycount, =@nextline
			rdword	addr, =@backbuf
			mov	t1, ycount
			add	t1, #1
			wrword	t1, =@nextline
			lockclr	zero

			cmp	ycount, #100		wc wz
		if_b	jmp	#:doline

			' The frame is complete. Wait for the next one.

:wait			lockset	zero			wc
		if_c	jmp	#:wait
			rdword	ycount, =@nextline
			rdword	addr, =@backbuf
			cmp	ycount, #100		wc wz
		if_b	mov	t1, ycount
		if_b	add	t1, #1
		if_b	wrword	t1, =@nextline
			lockclr	zero
		if_ae	jmp	#:wait

			' New frame.

			add	frame, #1

			add	angleanim, =$0020

			sub	zoom, =9
			cmps	zoom, =30 << 4		wc wz
		if_b	sub	zoom, =8
			mins	zoom, #0

			mov	sin_a, angleanim
			call	#sin
			mov	mulu_x, sin_res
			mov	mulu_s, zoom
			shr	mulu_s, #4
			call	#mulu
			negnz	scaled_sin, mulu_s

			call	#cos
			mov	mulu_x, sin_res
			mov	mulu_s, zoom
			shr	mulu_s, #4
			call	#mulu
			negnz	scaled_cos, mulu_s

			' Frame-global parameters: scaled_sin and scaled_cos
			' are used when mapping from object coordinates on the plane
			' to mandelbrot seeds. They are fixed point 8.24 (signed) in ]-1, 1[.

			' Draw a line.
:doline
			mov	sin_a, frame
			shl	sin_a, #6
			max	sin_a, =$7ff
			call	#sin
			shr	sin_res, #10

			mov	t1, ycount
			sub	t1, #50
			abs	t1, t1
			cmp	t1, sin_res		wc wz
		if_ae	jmp	#loop

			mov	t1, ycount
			shl	t1, #7
			add	addr, =128 * 116 + 28 - 8 * 128
			sub	addr, t1

			' Calculate the world y coordinate.

			mov	divu_n, ycount
			sub	divu_n, #48			' ydisp as 25.7
			mov	divu_d, =$8000			' 1 as 17.15
			shl	divu_n, #8
			sub	divu_d, divu_n
			shl	divu_n, #15			' ydisp as 2.30 (well within [-.5, .5])
			test	divu_n, =$80000000	wz
			abs	divu_n, divu_n
			call	#divu
			negnz	yobj, divu_n			' yobj as 17.15 (in [-.83, .83])

			' Calculate the texture coordinates for xworld = 0.
			' smidrow = scaled_sin * yobj + centerx
			' tmidrow = scaled_cos * yobj + centery

			mov	mulu_x, scaled_sin
			sar	mulu_x, #8
			mov	mulu_s, yobj
			sar	mulu_s, #7
			call	#muls				' result as 8.24
			shl	mulu_s, #6
			mov	rowmid_re, mulu_s

			mov	mulu_x, scaled_cos
			sar	mulu_x, #8
			mov	mulu_s, yobj
			sar	mulu_s, #7
			call	#muls				' result as 8.24
			shl	mulu_s, #6
			mov	rowmid_im, mulu_s

			' (will add centerx, centery later)

			' Calculate the difference in texture coordinates per xdisp.
			' ds/dxdisp = scaled_cos * (yobj + 1) = (tmidrow - centery) + scaled_cos
			' dt/dxdisp = -scaled_sin * (yobj + 1) = -(smidrow - centerx) - scaled_sin

			mov	xdelta_re, rowmid_im
			sar	xdelta_re, #5
			add	xdelta_re, scaled_cos
			sar	xdelta_re, #2

			neg	xdelta_im, rowmid_re
			sar	xdelta_im, #5
			sub	xdelta_im, scaled_sin
			sar	xdelta_im, #2

			add	rowmid_re, pivot_re
			add	rowmid_im, pivot_im

			' Find the texture coordinates at start of line.

			mov	pos_re, rowmid_re
			mov	pos_im, rowmid_im

			mov	t1, xdelta_re
			shl	t1, #5
			sub	pos_re, t1

			mov	t1, xdelta_im
			shl	t1, #5
			sub	pos_im, t1

			mov	xcount, #64
:xloop
			mov	pos_re_sh, pos_re
			sar	pos_re_sh, #16
			mov	pos_im_sh, pos_im
			sar	pos_im_sh, #16

			' Optimize away a block in the large cardioid, and one
			' in each of the two period-3 bulbs.
			
			cmps	pos_re_sh, =$00000900	wc wz
		if_ae	jmp	#:ok
			cmps	pos_re_sh, =$ffffee00	wc wz
		if_b	jmp	#:ok

			abs	t1, pos_im_sh
			cmp	t1, =$00001200		wc wz
		if_b	jmp	#:optimize

			cmp	t1, =$00001a80		wc wz
		if_ae	jmp	#:ok

			cmps	pos_re_sh, =$fffffe20	wc wz
		if_ae	jmp	#:ok

			cmps	pos_re_sh, =$fffff9e0	wc wz
		if_b	jmp	#:ok
:optimize
			mov	iterations, #0
			jmp	#:escape
:ok
			mov	z_re, #0
			mov	z_im, #0

			mov	iterations, =$2a / 2

			' 26, rd, 18, rd, 6, rd, 10, rd, 13, rd, 8
			' = 26 instructions per iteration
:iter
			' For each iteration, z += pos, check limits, z := z * z

			add	z_re, pos_re_sh
			add	z_im, pos_im_sh

			' Check limits by determining whether |re| or |im| > 2

			abs	t1, z_re
			cmp	t1, =2 << 13		wc wz
		if_be	abs	t1, z_im
		if_be	cmp	t1, =2 << 13		wc wz
		if_a	jmp	#:escape

			' Perform z := z * z
			' <=> z_re := z_re * z_re - z_im * z_im,
			'     z_im := 2 * z_re * z_im
			' <=> z_re := exp(2 * log(z_re)) - exp(2 * log(z_im)),
			'     z_im := 2 * exp(log(z_re) + log(z_im))

			abs	log_in, z_re		wc
			abs	log_in2, z_im

			mov	z_re_log, #0
			mov	z_im_log, #0

			test	log_in, =$ff00		wz
			muxnz	z_re_log, =$00080000
		if_z	shl	log_in, #8
			test	log_in, =$f000		wz
			muxnz	z_re_log, =$00040000
		if_z	shl	log_in, #4
			test	log_in, =$c000		wz
			muxnz	z_re_log, =$00020000
		if_z	shl	log_in, #2
			test	log_in, =$8000		wz
			muxnz	z_re_log, =$00010000
		if_z	shl	log_in, #1
			shr	log_in, #3
			and	log_in, =$ffe
			add	log_in, =$c000
			rdword	log_in, log_in
			or	z_re_log, log_in

			mov	exp_in, z_re_log
			add	exp_in, z_re_log

			test	log_in2, =$ff00		wz
			muxnz	z_im_log, =$00080000
		if_z	shl	log_in2, #8
			test	log_in2, =$f000		wz
			muxnz	z_im_log, =$00040000
		if_z	shl	log_in2, #4
			test	log_in2, =$c000		wz
			muxnz	z_im_log, =$00020000
		if_z	shl	log_in2, #2
			test	log_in2, =$8000		wz
			muxnz	z_im_log, =$00010000
		if_z	shl	log_in2, #1
			shr	log_in2, #3
			and	log_in2, =$ffe
			add	log_in2, =$c000
			rdword	log_in2, log_in2
			or	z_im_log, log_in2

			mov	exp_in2, z_im_log

			mov	exp_out, exp_in
			shr	exp_out, #4
			and	exp_out, =$ffe
			add	exp_out, =$d000
			rdword	z_re, exp_out
			shl	z_re, #15
			or	z_re, =$80000000
			shr	exp_in, #16
			xor	exp_in, #$1f
			shr	z_re, exp_in

			add	exp_in2, z_im_log

			mov	exp_out, exp_in2
			shr	exp_out, #4
			and	exp_out, =$ffe
			add	exp_out, =$d000
			rdword	exp_out, exp_out
			shl	exp_out, #15
			or	exp_out, =$80000000
			shr	exp_in2, #16
			xor	exp_in2, #$1f
			shr	exp_out, exp_in2

			sub	z_re, exp_out
			sar	z_re, #13

			test	z_im, =$80000000	wz
	
			add	z_re_log, z_im_log

			mov	exp_out, z_re_log
			shr	exp_out, #4
			and	exp_out, =$ffe
			add	exp_out, =$d000
			'nop
			rdword	z_im, exp_out
			shl	z_im, #15
			or	z_im, =$80000000
			shr	z_re_log, #16
			xor	z_re_log, #$1f
			shr	z_im, z_re_log

			shr	z_im, #12

	if_c_eq_z	neg	z_im, z_im

			djnz	iterations, #:iter
:escape
			shl	iterations, #1
			wrbyte	iterations, addr

			add	pos_re, xdelta_re
			add	pos_im, xdelta_im
			add	addr, #1
			djnz	xcount, #:xloop

			jmp	#loop

muls
			abs	mulu_x, mulu_x		wc
			rcl	signed, #1
			abs	mulu_s, mulu_s		wc
		if_c	xor	signed, #1
			call	#mulu
			rcr	signed, #1		wc
			negc	mulu_s, mulu_s
muls_ret		ret

			' mulu_x[15..0] * mulu_s[7..0] -> mulu_s[23..0] (clobbers c, preserves z)
			' 4*21 cycles
mulu
			shl	mulu_x, #16
			shr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
			shr	mulu_s, #8
mulu_ret		ret

divu			shl	divu_d, #15
			cmpsub	divu_n, divu_d		wc
			rcl	divu_n, #1
			cmpsub	divu_n, divu_d		wc
			rcl	divu_n, #1
			cmpsub	divu_n, divu_d		wc
			rcl	divu_n, #1
			cmpsub	divu_n, divu_d		wc
			rcl	divu_n, #1
			cmpsub	divu_n, divu_d		wc
			rcl	divu_n, #1
			cmpsub	divu_n, divu_d		wc
			rcl	divu_n, #1
			cmpsub	divu_n, divu_d		wc
			rcl	divu_n, #1
			cmpsub	divu_n, divu_d		wc
			rcl	divu_n, #1
			cmpsub	divu_n, divu_d		wc
			rcl	divu_n, #1
			cmpsub	divu_n, divu_d		wc
			rcl	divu_n, #1
			cmpsub	divu_n, divu_d		wc
			rcl	divu_n, #1
			cmpsub	divu_n, divu_d		wc
			rcl	divu_n, #1
			cmpsub	divu_n, divu_d		wc
			rcl	divu_n, #1
			cmpsub	divu_n, divu_d		wc
			rcl	divu_n, #1
			cmpsub	divu_n, divu_d		wc
			rcl	divu_n, #1
			cmpsub	divu_n, divu_d		wc
			rcl	divu_n, #1
			and	divu_n, =$ffff
divu_ret		ret

sincos
			call	#sin
			negnz	sin_sin, sin_res
			call	#cos
			negnz	sin_cos, sin_res
sincos_ret		ret

			' |sin(sin_a[12..0])| -> sin_a[15..0], (sin(sin_a[12..0]) >= 0) -> z

cos			add	sin_a, =$0800
sin			mov	sin_addr, sin_a
			test	sin_addr, =$0800	wc
			test	sin_addr, =$1000	wz
			and	sin_addr, =$07ff
		if_c	xor	sin_addr, =$07ff
			shl	sin_addr, #1
			or	sin_addr, =$e000
			rdword	sin_res, sin_addr
cos_ret
sin_ret			ret

xdelta_re		long	$00000000
xdelta_im		long	$00800000

zoom			long	200 << 4

frame			long	0

			pool

pivot_re		long	882467	
pivot_im		long	441558948
angleanim		long	3 << 11

t1			res	1
addr			res	1
iterations		res	1
xcount			res	1
ycount			res	1
accum			res	1

z_re			res	1
z_im			res	1
pos_re			res	1
pos_im			res	1
pos_re_sh		res	1
pos_im_sh		res	1
rowmid_re		res	1
rowmid_im		res	1

z_re_log		res	1
z_im_log		res	1

exp_in			res	1
exp_in2			res	1
exp_out			res	1

log_in			res	1
log_in2			res	1

scaled_sin		res	1
scaled_cos		res	1

sin_a			res	1
sin_addr		res	1
sin_res			res	1
sin_sin			res	1
sin_cos			res	1

mulu_s			res	1
mulu_x			res	1
signed			res	1

divu_d			res	1
divu_n			res	1

yobj			res	1

			vfit	496
