
			include "propeller.i"
			include "demo.i"

			org	0

			include	"unpack.i"
			include	"eeprom.i"

			' Before this code is the unpacker with the EEPROM routine plugged in.
			' We'll make calls to the unpacker and the bitstream reader.

			' Ensure that any unused palette bits are held low.

			or	DIRA, #$ff

			' Locate the filetable

			rdword	filetable, =@filetableptr

			' Now that we're past the initialization stage, we're finally free to trash hub
			' RAM below $7400.

			' Set up the bitfile pointers

			mov	fileno, #BF_SONG
			call	#seek
			mov	song, filereader
			mov	restart, song

			mov	fileno, #BF_TRACKS
			call	#seek
			mov	trackfile, filereader

			mov	fileno, #BF_INSTRUMENTS
			call	#seek
			mov	instrfile, filereader

			mov	fileno, #BF_SCRIPT
			call	#seek
			mov	scriptreader, filereader

			' Prepare the 60 Hz timer

			mov	timer, CNT
			add	timer, deltacnt

			' Patch the unpacker to call playroutine regularly.

			movs	p_unpack, #hook

			' Main loop - parse the script, and call playroutine at least every 1000 us

main_re
			mov	reader, scriptreader
main
			mov	gc_count, #3
			call	#getchunk
			add	gc_result, #script_table
			jmp	gc_result

script_table
			jmp	#scr_wait
			jmp	#scr_seek
			jmp	#scr_unp_file
			jmp	#scr_unp_script
			jmp	#scr_launch
			jmp	#scr_stop
			jmp	#scr_reset
			'jmp	#scr_clear

scr_clear
			mov	addr, #0
			mov	count, =@unpack_area / 4
:loop			wrlong	zero, addr
			add	addr, #4
			djnz	count, #:loop
			wrbyte	zero, =@ditherstyle
			' drop through
scr_reset
			mov	t1, #511
			wrword	t1, =@nextline
			wrbyte	zero, =@slave_enable
			lockclr	zero
			jmp	#main

scr_stop
			mov	gc_count, #3
			call	#getchunk
			cogstop	gc_result
			jmp	#main

scr_launch
			mov	gc_count, #3
			call	#getchunk
			or	gc_result, =@unpack_area << 2
			call	#getbit
			muxc	gc_result, =1 << 18
			coginit	gc_result
			jmp	#main

scr_unp_script
			mov	up_dest, =@palette_area - 1
			call	#unpack
			jmp	#main

scr_unp_file
			mov	gc_count, #2
			call	#getchunk
			add	:p_dest, gc_result
			sub	:p_dest, gc_result	' goes into pipeline slot
:p_dest			mov	up_dest, dest_table
			mov	scriptreader, reader
			mov	reader, filereader
			call	#unpack
			mov	filereader, reader
			jmp	#main_re

dest_table		long	$ffffffff
			long	@unpack_area - 1
			long	@extra_area - 1

scr_seek
			mov	gc_count, #5
			call	#getchunk
			mov	fileno, gc_result
			mov	scriptreader, reader
			call	#seek
			jmp	#main_re

scr_wait
			mov	gc_count, #10
			call	#getchunk
			tjz	gc_result, #endofscript
			shl	gc_result, #2		' multiply by TEMPO
			add	scriptclock, gc_result
:wait
			cmp	playclock, scriptclock	wc wz
		if_ae	jmp	#main
			call	#playroutine
			jmp	#:wait

			' A "call #getbit" in the unpacker gets redirected here. The
			' return address (p_unpack_ret) will obviously be in getbit_ret.
			' However, the playroutine can make calls to
			' getbit (but not unpack), so we need to write the return address
			' back into getbit_ret before handing over control.
hook
			call	#playroutine
			movs	getbit_ret, #p_unpack_ret
			jmp	#getbit

endofscript
			movs	playroutine_ret, #playroutine	' repeatedly run the playroutine

playroutine
			' Time to run yet? If more than 1000 us left, just return.

			mov	t1, timer
			sub	t1, CNT
			cmp	t1, margin		wc wz
		if_a	jmp	playroutine_ret

			' Synchronize with the 60 Hz timer

			waitcnt	timer, deltacnt

			add	playclock, #1
			mov	hooksave, reader

			djnz	tracktimer, #nonewline
			mov	tracktimer, #TEMPO

			' Time for a new track line

			djnz	tracklines, #nonewpos
			mov	tracklines, #TRACKLEN

			' Time for a new song line

			mov	addr, =@channels+CH_LASTT
			mov	count, #6
			mov	reader, song
:lp
			call	#getbit
			mov	gc_count, #TRACKBITS
			rdbyte	gc_result, addr
		if_c	call	#getchunk
			wrbyte	gc_result, addr

			mov	savereader, reader
			mov	reader, gc_result
			add	reader, reader
			add	reader, trackfile
			mov	gc_count, #16
			call	#getchunk
			rev	gc_result, #16
			mov	t1, gc_result
			mov	reader, savereader

			sub	addr, #CH_LASTT-CH_TRACK
			wrlong	t1, addr

			call	#getbit
			mov	gc_count, #TRANSPBITS
			mov	gc_result, #0
		if_c	call	#getchunk
			shl	gc_result, #32-TRANSPBITS	wz
			sar	gc_result, #32-TRANSPBITS

		if_nz	tjz	t1, #nonewline		' A transposed 00-track signals end of song

			add	addr, #CH_TRANSP-CH_TRACK
			wrbyte	gc_result, addr
			add	addr, #CH_SIZEOF-CH_TRANSP+CH_LASTT
			djnz	count, #:lp

			mov	song, reader

			' All track bit readers are valid at this point.
nonewpos
			mov	addr, =@channels+CH_TRACK
			mov	count, #6
:lp
			rdlong	reader, addr
			tjz	reader, #:notrack

			call	#getbit
		if_nc	jmp	#:noinstr

			mov	gc_count, #INSTRBITS
			call	#getchunk

			mov	inum, gc_result			wz
		if_z	mov	reader, #0
		if_z	jmp	#:nonote

			add	addr, #CH_INSTR-CH_TRACK
			call	#gotoinstr

			add	addr, #CH_TIMER-CH_INSTR
			wrbyte	zero, addr
			add	addr, #CH_LASTINSTR-CH_TIMER
			wrbyte	inum, addr
			sub	addr, #CH_LASTINSTR-CH_TRACK
:noinstr
			call	#getbit
		if_nc	jmp	#:nonote

			mov	gc_count, #NOTEBITS
			call	#getchunk

			sub	addr, #CH_TRACK-CH_WAVE
			wrbyte	zero, addr

			add	addr, #CH_TRANSP-CH_WAVE
			rdbyte	inum, addr
			add	gc_result, inum
			and	gc_result, #$ff
			movd	gc_result, #12			' This becomes the default vibrato rate (24)
			shl	gc_result, #8
			sub	addr, #CH_TRANSP-CH_INOTE
			wrword	gc_result, addr
			add	addr, #CH_TIMER-CH_INOTE
			wrlong	zero, addr
			add	addr, #CH_LASTINSTR-CH_TIMER
			rdbyte	inum, addr
			or	gc_result, inum
			wrlong	gc_result, addr
			sub	addr, #CH_LASTINSTR-CH_INSTR
			call	#gotoinstr
			sub	addr, #CH_INSTR-CH_TRACK
:nonote
			wrlong	reader, addr
:notrack
			add	addr, #CH_SIZEOF
			djnz	count, #:lp
nonewline
			' Run the instruments and update the synth parameters

			mov	addr, =@channels
			mov	count, #6
chloop
			rdlong	sound, addr

			' Run the instrument

			add	addr, #CH_TIMER-CH_WAVE
			mov	saveaddr, addr
runcmd_done
			rdbyte	t1, saveaddr
			tjnz	t1, #:waiting

			mov	addr, saveaddr
			sub	addr, #CH_TIMER-CH_INSTR
			rdlong	reader, addr
			tjz	reader, #:noinstr

			mov	gc_count, #4
			call	#getchunk
			add	gc_result, gc_result
			add	gc_result, #runcmd_table
			movs	:p_cmd, gc_result
			mov	gc_count, #8
			call	#getchunk
			wrlong	reader, addr

:p_cmd			jmp	#000

:waiting		sub	t1, #1
			wrbyte	t1, saveaddr
:noinstr
			mov	addr, saveaddr		' addr points to CH_TIMER

			' Update the parameters

			add	addr, #CH_SLIDE-CH_TIMER
			call	#rdsbyte
			shl	t1, #3

			sub	addr, #CH_SLIDE-CH_INOTE
			rdword	freq, addr

			add	t1, freq
			test	t1, =$8000		wz
		if_nz	mov	t1, #0
			wrword	t1, addr

			add	addr, #CH_VRATE-CH_INOTE
			rdbyte	t1, addr

			add	addr, #CH_VPOS-CH_VRATE
			rdbyte	sin_a, addr

			add	t1, sin_a
			wrbyte	t1, addr

			shl	sin_a, #5
			call	#sin

			sub	addr, #CH_VPOS-CH_VDEPTH
			rdbyte	mulu_s, addr
			call	#mulu

			negnz	mulu_s, mulu_s
			sar	mulu_s, #13
			add	mulu_s, freq

			mov	t1, mulu_s
			shr	t1, #7
			and	t1, #$fe
			add	t1, =@freqtable
			rdword	freq, t1
			add	t1, #2
			rdword	mulu_x, t1
			sub	mulu_x, freq

			call	#mulu

			shr	mulu_s, #8
			add	freq, mulu_s

			sub	addr, #CH_VDEPTH-CH_SWEEP
			call	#rdsbyte
			mov	mulu_s, sound
			shr	mulu_s, #9		' amount
			and	mulu_s, #$ff
			add	mulu_s, t1
			mins	mulu_s, #0
			max	mulu_s, #$ff		wc
			negnc	t1, t1
			wrbyte	t1, addr
			movd	sound, mulu_s

			sub	addr, #CH_SWEEP-CH_FADE
			call	#rdsbyte
			movs	b1, sound
			add	t1, b1
			mins	t1, #0
			max	t1, #$ff
			movs	sound, t1

			' Write sound & freq

			sub	addr, #CH_FADE-CH_WAVE
			wrlong	sound, addr
			add	addr, #CH_FREQ-CH_WAVE
			wrword	freq, addr
			add	addr, #CH_SIZEOF-CH_FREQ
			djnz	count, #chloop

			mov	reader, hooksave

playroutine_ret		ret

gotoinstr
			mov	savereader, reader
			mov	reader, instrfile
			add	reader, inum
			mov	gc_count, #8
			call	#getchunk
			rev	gc_result, #24
			add	gc_result, gc_result
			add	gc_result, instrfile
			wrlong	gc_result, addr
			mov	reader, savereader
gotoinstr_ret		ret

rdsbyte
			rdbyte	t1, addr
			shl	t1, #24
			sar	t1, #24
rdsbyte_ret		ret

			' addr points to CH_INSTR
runcmd_table
			wrlong	zero, addr			' 0
			jmp	#runcmd_done

			mov	inum, gc_result			' 1
			call	#gotoinstr
			jmp	#runcmd_done
scriptclock		long	0			' Hole in jump table, used as variable.

			add	addr, #CH_SLIDE-CH_INSTR	' 3
			jmp	#runcmd_wr

			add	addr, #CH_VRATE-CH_INSTR	' 4
			jmp	#runcmd_wr

			add	addr, #CH_VDEPTH-CH_INSTR	' 5
			jmp	#runcmd_wr

			movs	sound, gc_result		' 6
			jmp	#runcmd_done

			add	addr, #CH_FADE-CH_INSTR		' 7
			jmp	#runcmd_wr

			movd	sound, gc_result		' 8
			jmp	#runcmd_done

			mov	song, restart			' 9
			jmp	#runcmd_done

			add	addr, #CH_TNOTE-CH_INSTR	' a
			jmp	#runcmd_a

			movi	sound, gc_result		' b
			jmp	#runcmd_done

			add	addr, #CH_SWEEP-CH_INSTR	' c
			jmp	#runcmd_wr

runcmd_d		add	addr, #CH_TIMER-CH_INSTR	' d

runcmd_wr		wrbyte	gc_result, addr
			jmp	#runcmd_done

runcmd_a		rdbyte	t1, addr
			add	gc_result, t1
			sub	addr, #CH_TNOTE-(CH_INOTE+1)
			jmp	#runcmd_wr

			' mulu_x[15..0] * mulu_s[7..0] -> mulu_s[23..0] (clobbers c, preserves z)
			' 4*21 cycles
mulu
			shl	mulu_x, #16
			shr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
			shr	mulu_s, #8
mulu_ret		ret

			' |sin(sin_a[12..0])| -> sin_a[15..0], (sin(sin_a[12..0]) >= 0) -> z

sin			test	sin_a, =$0800		wc
			test	sin_a, =$1000		wz
			and	sin_a, =$07ff
		if_c	xor	sin_a, =$07ff
			shl	sin_a, #1
			or	sin_a, =$e000
			rdword	mulu_x, sin_a
sin_ret			ret

seek
			mov	reader, filetable
			add	reader, fileno
			add	reader, fileno
			mov	gc_count, #16
			call	#getchunk
			rev	gc_result, #16
			mov	filereader, gc_result
seek_ret		ret

deltacnt		long	80000000/60
margin			long	80*1000

tracktimer		long	1
tracklines		long	1

			pool

b1			long	0
playclock		long	0

count			res	1
inum			res	1
addr			res	1
timer			res	1

sound			res	1
freq			res	1

song			res	1
restart			res	1
instrfile		res	1
trackfile		res	1

sin_a			res	1
mulu_s			res	1
mulu_x			res	1

saveaddr		res	1
savereader		res	1
hooksave		res	1

filetable		res	1
fileno			res	1
filereader		res	1
scriptreader		res	1

			vfit	496
